# coding:utf-8
# Copyright (c) 2022  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class BaseRouterManager(abc.ABC):
    _app = None

    def __init__(self, app):
        super().__init__()
        self._app = app

    @abc.abstractmethod
    def register_models_router(self):
        return NotImplemented

    @abc.abstractmethod
    def register_taskflow_router(self):
        return NotImplemented
